/*
 *  This file is autogenerated from evolution-message-composer.xml, do not edit
 *
 *  This file contains translatable strings generated by
 * bonobo-ui-extract. Add this file to your project's
 * POTFILES.in. DO NOT compile it as part of your application.
 */

gchar *s = N_("Open");
gchar *s = N_("Open a file");
gchar *s = N_("Save");
gchar *s = N_("Save the current file");
gchar *s = N_("Save As");
gchar *s = N_("Save the current file with a different name");
gchar *s = N_("Close");
gchar *s = N_("Close the current file");
gchar *s = N_("Save in folder...");
gchar *s = N_("Save the message in a specified folder");
gchar *s = N_("Insert text file...");
gchar *s = N_("Insert a file as text into the message");
gchar *s = N_("Send");
gchar *s = N_("Send the message now");
gchar *s = N_("Send _later");
gchar *s = N_("Send the message later");
gchar *s = N_("Attach");
gchar *s = N_("Attach a file");
gchar *s = N_("HTML");
gchar *s = N_("Send the mail in HTML format");
gchar *s = N_("Show attachments");
gchar *s = N_("Show / hide attachments");
gchar *s = N_("_File");
gchar *s = N_("_Open...");
gchar *s = N_("_Save");
gchar *s = N_("Save _As...");
gchar *s = N_("Save in _folder... (FIXME)");
gchar *s = N_("_Insert text file... (FIXME)");
gchar *s = N_("Send");
gchar *s = N_("Send _Later");
gchar *s = N_("_Close");
gchar *s = N_("_Edit");
gchar *s = N_("F_ormat");
gchar *s = N_("_View");
gchar *s = N_("Show _attachments");
gchar *s = N_("_Help");
gchar *s = N_("_About...");
gchar *s = N_("_Debug");
gchar *s = N_("Send this message now");
gchar *s = N_("Bold");
gchar *s = N_("Sets something as bold");




