/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*- */
/* e-select-names-bonobo.h
 *
 * Copyright (C) 2000  Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Author: Ettore Perazzoli
 */

#ifndef __E_SELECT_NAMES_BONOBO_H__
#define __E_SELECT_NAMES_BONOBO_H__

#include <bonobo/bonobo-object.h>

#include "Evolution-Addressbook-SelectNames.h"

#ifdef __cplusplus
extern "C" {
#pragma }
#endif /* __cplusplus */

#define E_TYPE_SELECT_NAMES_BONOBO			(e_select_names_bonobo_get_type ())
#define E_SELECT_NAMES_BONOBO(obj)			(G_TYPE_CHECK_INSTANCE_CAST ((obj), E_TYPE_SELECT_NAMES_BONOBO, ESelectNamesBonobo))
#define E_SELECT_NAMES_BONOBO_CLASS(klass)		(G_TYPE_CHECK_CLASS_CAST ((klass), E_TYPE_SELECT_NAMES_BONOBO, ESelectNamesBonoboClass))
#define E_IS_SELECT_NAMES_BONOBO(obj)			(G_TYPE_CHECK_INSTANCE_TYPE ((obj), E_TYPE_SELECT_NAMES_BONOBO))
#define E_IS_SELECT_NAMES_BONOBO_CLASS(klass)		(G_TYPE_CHECK_CLASS_TYPE ((klass), E_TYPE_SELECT_NAMES_BONOBO))


typedef struct _ESelectNamesBonobo        ESelectNamesBonobo;
typedef struct _ESelectNamesBonoboPrivate ESelectNamesBonoboPrivate;
typedef struct _ESelectNamesBonoboClass   ESelectNamesBonoboClass;

struct _ESelectNamesBonobo {
	BonoboObject parent;

	ESelectNamesBonoboPrivate *priv;
};

struct _ESelectNamesBonoboClass {
	BonoboObjectClass parent_class;

	POA_GNOME_Evolution_Addressbook_SelectNames__epv epv;
};


GType               e_select_names_bonobo_get_type   (void);
ESelectNamesBonobo *e_select_names_bonobo_new        (void);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __E_SELECT_NAMES_BONOBO_H__ */
