/* -*- Mode: IDL; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Interface for a wizard done through Bonobo.
 *
 * Authors:
 *	Iain Holmes  <iain@ximian.com>
 *
 * Copyright (C) 2000, 2001 Ximian, Inc.
 */

#include <Bonobo.idl>

module GNOME {
module Evolution {

	interface Wizard : Bonobo::Unknown {
		
		struct Page {
			string          title;
			Icon            icon;
			Bonobo::Control control;
		};
		typedef sequence<Page> PageList;

		readonly attribute PageList pages;

		enum Action {
			NEXT,
			PREPARE,
			BACK,
			FINISH,
			CANCEL,
			HELP
		};

		exception NoPage {};

		void notifyAction (in long pagenumber, in Action action)
			raises (NoPage);
	};
};
};
