/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/* 
 * Copyright 2000 Ximian, Inc. (www.ximian.com)
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of version 2 of the GNU General Public 
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#ifndef MAIL_TYPES_H
#define MAIL_TYPES_H 1

#ifdef __cplusplus
extern "C" {
#pragma }
#endif /* __cplusplus }*/


typedef struct  _FolderBrowser     FolderBrowser;
typedef struct  _MessageBrowser    MessageBrowser;
typedef struct  _SubscribeDialog   SubscribeDialog;
typedef struct  _MessageList       MessageList;
typedef struct  _MailDisplay       MailDisplay;

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* MAIL_TYPES_H */
