/*
 *  This file is autogenerated from evolution-addressbook.xml, do not edit
 *
 *  This file contains translatable strings generated by
 * bonobo-ui-extract. Add this file to your project's
 * POTFILES.in. DO NOT compile it as part of your application.
 */

gchar *s = N_("_Print Contacts...");
gchar *s = N_("As _Table");
gchar *s = N_("_New Contact");
gchar *s = N_("_Tools");
gchar *s = N_("_Search for contacts");
gchar *s = N_("New");
gchar *s = N_("Create a new contact");
gchar *s = N_("Find");
gchar *s = N_("Find a contact");
gchar *s = N_("Print");
gchar *s = N_("Print contacts");
gchar *s = N_("Delete");
gchar *s = N_("Delete a contact");
gchar *s = N_("View All");
gchar *s = N_("View all contacts");
gchar *s = N_("Stop");
gchar *s = N_("Stop Loading");
