/* WARNING: Exceptions MUST NOT be renumbered: they need to be
 * consistent across libraries compiled at different times.
 * Categories should be widely separated, old unused exceptions can
 * never be deleted, and new exceptions can be added only to the
 * ends of categories.
 */

CAMEL_EXCEPTION_NONE = 0,

/* Generic exceptions */
CAMEL_EXCEPTION_INVALID_PARAM,
CAMEL_EXCEPTION_SYSTEM,
CAMEL_EXCEPTION_USER_CANCEL,

/* CamelFolderException */
CAMEL_EXCEPTION_FOLDER_NULL = 100,
CAMEL_EXCEPTION_FOLDER_INVALID,
CAMEL_EXCEPTION_FOLDER_INVALID_STATE,
CAMEL_EXCEPTION_FOLDER_NON_EMPTY,
CAMEL_EXCEPTION_FOLDER_NON_UID,
CAMEL_EXCEPTION_FOLDER_INSUFFICIENT_PERMISSION,
CAMEL_EXCEPTION_FOLDER_INVALID_PATH,
CAMEL_EXCEPTION_FOLDER_INVALID_UID,
CAMEL_EXCEPTION_FOLDER_SUMMARY_INVALID,

/* CamelStoreException */
CAMEL_EXCEPTION_STORE_NULL = 200,
CAMEL_EXCEPTION_STORE_INVALID,
CAMEL_EXCEPTION_STORE_NO_FOLDER,

/* CamelServiceException */
CAMEL_EXCEPTION_SERVICE_NULL = 300,
CAMEL_EXCEPTION_SERVICE_INVALID,
CAMEL_EXCEPTION_SERVICE_URL_INVALID,
CAMEL_EXCEPTION_SERVICE_UNAVAILABLE,
CAMEL_EXCEPTION_SERVICE_CANT_AUTHENTICATE,
CAMEL_EXCEPTION_SERVICE_NOT_CONNECTED
